@@ -59,8 +59,7 @@ ways. Some routines have no way to return an error (eg `xDlOpen
 <http://www.sqlite.org/c3ref/vfs.html>`_ just returns zero/NULL on
 being unable to load a library, `xSleep
 <http://www.sqlite.org/c3ref/vfs.html>`_ has no error return
-parameter), others have any error values ignored (eg
-:cvstrac:`xCurrentTime <3394>`), others are unified (eg almost any
+parameter), others are unified (eg almost any
 error in xWrite will be returned to the user as disk full
 error). Sometimes errors are ignored as they are harmless such as when
 a journal can't be deleted after a commit (the journal is marked as
@@ -222,16 +221,24 @@ typedef struct
   PyObject_HEAD
   struct sqlite3_file *base;
   char *filename;  /* obtained from fullpathname - has to be around for lifetime of base */
+  int filenamefree;  /* filename should be freed on close */
   /* If you add any new members then also initialize them in
      apswvfspy_xOpen() as that function does not call init because it
      has values already */
 } APSWVFSFile;
 
 static PyTypeObject APSWVFSFileType;
+static PyTypeObject APSWURIFilenameType;
 
 static const struct sqlite3_io_methods apsw_io_methods_v1;
 static const struct sqlite3_io_methods apsw_io_methods_v2;
 
+typedef struct
+{
+  PyObject_HEAD
+  char *filename;
+} APSWURIFilename;
+
 
 /** .. class:: VFS
 
@@ -292,7 +299,13 @@ apswvfs_xDelete(sqlite3_vfs *vfs, const char *zName, int syncDir)
 
 /** .. method:: xDelete(filename, syncdir)
 
-    Delete the named file.
+    Delete the named file.  
+
+    .. note::
+
+       SQLite 3.7.8 changed semantics if the file does not exist.  In
+       prior SQLite versions an error should be raised.  For SQLite
+       3.7.8 onwards you should not raise an error.
 
     :param filename: File to delete
 
@@ -481,6 +494,8 @@ apswvfs_xOpen(sqlite3_vfs *vfs, const char *zName, sqlite3_file *file, int infla
   PyObject *flags=NULL;
   PyObject *pyresult=NULL;
   APSWSQLite3File *apswfile=(APSWSQLite3File*)(void*)file;
+  /* how we pass the name */
+  PyObject *nameobject;
 
   VFSPREAMBLE;
 
@@ -491,7 +506,16 @@ apswvfs_xOpen(sqlite3_vfs *vfs, const char *zName, sqlite3_file *file, int infla
   PyList_SET_ITEM(flags, 1, PyInt_FromLong(pOutFlags?*pOutFlags:0));
   if(PyErr_Occurred()) goto finally;
 
-  pyresult=Call_PythonMethodV((PyObject*)(vfs->pAppData), "xOpen", 1, "(NO)", convertutf8string(zName), flags);
+  if(inflags & (SQLITE_OPEN_URI|SQLITE_OPEN_MAIN_DB)) 
+    {
+      nameobject=PyObject_New(PyObject, &APSWURIFilenameType);
+      if(nameobject)
+	((APSWURIFilename*)nameobject)->filename=(char*)zName;
+    }
+  else
+    nameobject=convertutf8string(zName);
+
+  pyresult=Call_PythonMethodV((PyObject*)(vfs->pAppData), "xOpen", 1, "(NO)", nameobject, flags);
   if(!pyresult)
     {
       result=MakeSqliteMsgFromPyException(NULL);
@@ -547,11 +571,11 @@ apswvfs_xOpen(sqlite3_vfs *vfs, const char *zName, sqlite3_file *file, int infla
 
     :param name: File to open.  Note that *name* may be :const:`None` in which
         case you should open a temporary file with a name of your
-        choosing.
+        choosing.  May be an instance of :class:`URIFilename`.
 
     :param flags: A list of two integers ``[inputflags,
       outputflags]``.  Each integer is one or more of the `open flags
-      <http://www.sqlite.org/c3ref/c_open_create.html>`_ binary orred
+      <http://www.sqlite.org/c3ref/c_open_autoproxy.html>`_ binary orred
       together.  The ``inputflags`` tells you what SQLite wants.  For
       example :const:`SQLITE_OPEN_DELETEONCLOSE` means the file should
       be automatically deleted when closed.  The ``outputflags``
@@ -581,13 +605,28 @@ apswvfspy_xOpen(APSWVFS *self, PyObject *args)
 
   if(pyname==Py_None)
     {
-      utf8name=Py_None;
-      Py_INCREF(Py_None);
+      filename=NULL;
+    }
+  else if(pyname->ob_type==&APSWURIFilenameType)
+    {
+      filename=((APSWURIFilename*)pyname)->filename;
     }
   else
-    utf8name=getutf8string(pyname);
-  if(!utf8name) 
-    goto finally;
+    {
+      size_t len;
+      utf8name=getutf8string(pyname);
+      if(!utf8name)
+	goto finally;
+      len=strlen(PyBytes_AS_STRING(utf8name));
+      APSW_FAULT_INJECT(vfspyopen_fullpathnamemallocfailed,
+			filename=PyMem_Malloc(len+3),
+			filename=(char*)PyErr_NoMemory());
+      if(!filename) 
+	goto finally;
+      strcpy(filename, PyBytes_AS_STRING(utf8name));
+      /* ensure extra null padding for URI params */
+      filename[len]=filename[len+1]=filename[len+2]=0;
+    }
 
 
   if(!PyList_Check(flags) || PyList_GET_SIZE(flags)!=2 || !PyIntLong_Check(PyList_GET_ITEM(flags, 0)) || !PyIntLong_Check(PyList_GET_ITEM(flags, 1)))
@@ -606,27 +645,6 @@ apswvfspy_xOpen(APSWVFS *self, PyObject *args)
   file=PyMem_Malloc(self->basevfs->szOsFile);
   if(!file) goto finally;
 
-  if(utf8name!=Py_None)
-    {
-      /* The filename has to be passed through xFullPathname.  That
-	 may or may not already have happened, but we do it a second
-	 time to be sure. */
-      int fpres;
-      APSW_FAULT_INJECT(vfspyopen_fullpathnamemallocfailed,
-			filename=PyMem_Malloc(self->basevfs->mxPathname+1),
-			filename=(char*)PyErr_NoMemory());
-      if(!filename)
-	goto finally;
-      APSW_FAULT_INJECT(vfspyopen_fullpathnamefailed,
-			fpres=self->basevfs->xFullPathname(self->basevfs, PyBytes_AS_STRING(utf8name), self->basevfs->mxPathname, filename),
-			fpres=SQLITE_NOMEM);
-      if(fpres!=SQLITE_OK)
-	{
-	  SET_EXC(fpres, NULL);
-	  goto finally;
-	}
-    }
-
   res=self->basevfs->xOpen(self->basevfs, filename, file, flagsin, &flagsout);
   if(PyErr_Occurred()) goto finally;
   if(res!=SQLITE_OK)
@@ -642,13 +660,14 @@ apswvfspy_xOpen(APSWVFS *self, PyObject *args)
   if(!apswfile) goto finally;
   apswfile->base=file;
   apswfile->filename=filename;
+  apswfile->filenamefree=!!utf8name;
   filename=NULL;
   file=NULL;
   result=(PyObject*)(void*)apswfile;
 
  finally:
   if(file) PyMem_Free(file);
-  if(filename) PyMem_Free(filename);
+  if(utf8name && filename) PyMem_Free(filename);
   Py_XDECREF(utf8name);
   return result;
 }
@@ -878,10 +897,9 @@ apswvfs_xDlError(sqlite3_vfs *vfs, int nByte, char *zErrMsg)
     zero/NULL). If you do not supply this routine then SQLite provides
     a generic message. To implement this method, catch exceptions in
     :meth:`~VFS.xDlOpen` or :meth:`~VFS.xDlSym`, turn them into
-    strings, save them, and return them in this routine. Note that the
-    message may be truncated to 255 characters - see
-    :cvstrac:`3305`. If you have an error in this routine or return
-    None then SQLite's generic message will be used.  
+    strings, save them, and return them in this routine.  If you have
+    an error in this routine or return None then SQLite's generic
+    message will be used.
 */
 static PyObject *
 apswvfspy_xDlError(APSWVFS *self)
@@ -1067,12 +1085,10 @@ apswvfspy_xSleep(APSWVFS *self, PyObject *args)
   return PyLong_FromLong(self->basevfs->xSleep(self->basevfs, microseconds));
 }
 
-/* See :cvstrac:`3394` for SQLite implementation issues */
 static int
 apswvfs_xCurrentTime(sqlite3_vfs *vfs, double *julian)
 {
   PyObject *pyresult=NULL;
-  /* note returns zero or one.  Details in sqlite ticket 3394*/
   int result=0; 
   VFSPREAMBLE;
 
@@ -1098,8 +1114,7 @@ apswvfs_xCurrentTime(sqlite3_vfs *vfs, double *julian)
   <http://en.wikipedia.org/wiki/Julian_day>`_ as a floating point
   number where the integer portion is the day and the fractional part
   is the time. Do not adjust for timezone (ie use `UTC
-  <http://en.wikipedia.org/wiki/Universal_Time>`_). Although SQLite
-  allows for an error return, that is :cvstrac:`ignored <3394>`.
+  <http://en.wikipedia.org/wiki/Universal_Time>`_).
 */
 static PyObject *
 apswvfspy_xCurrentTime(APSWVFS *self)
@@ -1116,7 +1131,6 @@ apswvfspy_xCurrentTime(APSWVFS *self)
 
   if(res!=0)
     {
-      /* routines are documented to return zero or one - see ticket 3394 info above */
       SET_EXC(SQLITE_ERROR, NULL);   /* general sqlite error code */
       AddTraceBackHere(__FILE__, __LINE__, "vfspy.xCurrentTime", NULL);
       return NULL;
@@ -1539,18 +1553,18 @@ APSWVFS_new(PyTypeObject *type, APSW_ARGUNUSED PyObject *args, APSW_ARGUNUSED Py
 /** .. method:: __init__(name[, base=None, makedefault=False, maxpathname=1024])
 
     :param name: The name to register this vfs under.  If the name
-                 already exists then this vfs will replace the prior one of the
-                 same name.  Use :meth:`apsw.vfsnames` to get a list of
-                 registered vfs names.
-    :param base: If you would like to inherit behaviour from an
-                 already registered vfs then give their name.  To inherit from the
-                 default vfs, use a zero length string ``""`` as the name.
-    :param makedefault: If true then this vfs will be registered as
-      the default, and will be used by any opens that don't specify a
-      vfs.
-    :param maxpathname: The maximum length of database name in bytes
-      when represented in UTF-8.  If a pathname is passed in longer than
-      this value then SQLite will not :cvstrac:`be able to open it <3373>`.
+        already exists then this vfs will replace the prior one of the
+        same name.  Use :meth:`apsw.vfsnames` to get a list of
+        registered vfs names.  :param base: If you would like to
+        inherit behaviour from an already registered vfs then give
+        their name.  To inherit from the default vfs, use a zero
+        length string ``""`` as the name.  :param makedefault: If true
+        then this vfs will be registered as the default, and will be
+        used by any opens that don't specify a vfs.  :param
+        maxpathname: The maximum length of database name in bytes when
+        represented in UTF-8.  If a pathname is passed in longer than
+        this value then SQLite will not `be able to open it
+        <http://www.sqlite.org/src/tktview/c060923a5422590b3734eb92eae0c94934895b68>`__.
 
     :raises ValueError: If *base* is not :const:`None` and the named vfs is not
       currently registered.
@@ -1766,7 +1780,7 @@ APSWVFSFile_dealloc(APSWVFSFile *self)
       PyObject *x=apswvfsfilepy_xClose(self);
       Py_XDECREF(x);
     }
-  if(self->filename)
+  if(self->filenamefree)
     PyMem_Free(self->filename);
 
   if(PyErr_Occurred())
@@ -1798,7 +1812,7 @@ APSWVFSFile_new(PyTypeObject *type, APSW_ARGUNUSED PyObject *args, APSW_ARGUNUSE
 
     :param vfs: The vfs you want to inherit behaviour from.  You can
        use an empty string ``""`` to inherit from the default vfs.
-    :param name: The name of the file being opened.
+    :param name: The name of the file being opened.  May be an instance of :class:`URIFilename`.
     :param flags: A two list ``[inflags, outflags]`` as detailed in :meth:`VFS.xOpen`.
 
     :raises ValueError: If the named VFS is not registered.
@@ -1831,15 +1845,32 @@ APSWVFSFile_init(APSWVFSFile *self, PyObject *args, PyObject *kwds)
   if(!PyArg_ParseTupleAndKeywords(args, kwds, "esOO:init(vfs, name, flags)", kwlist, STRENCODING, &vfs, &pyname, &flags))
     return -1;
 
+  self->filenamefree=0;
   if(pyname==Py_None)
     {
-      utf8name=Py_None;
-      Py_INCREF(utf8name);
+      self->filename=NULL;
+    }
+  else if(pyname->ob_type==&APSWURIFilenameType)
+    {
+      self->filename=((APSWURIFilename*)pyname)->filename;
     }
   else
-    utf8name=getutf8string(pyname);
-
-  if(!utf8name) goto finally;
+    {
+      size_t len;
+      utf8name=getutf8string(pyname);
+      if(!utf8name)
+	goto finally;
+      len=strlen(PyBytes_AS_STRING(utf8name));
+      APSW_FAULT_INJECT(vfspyopen_fullpathnamemallocfailed,
+			self->filename=PyMem_Malloc(len+3),
+			self->filename=(char*)PyErr_NoMemory());
+      if(!self->filename) 
+	goto finally;
+      strcpy(self->filename, PyBytes_AS_STRING(utf8name));
+      /* ensure extra null padding for URI params */
+      self->filename[len]=self->filename[len+1]=self->filename[len+2]=0;
+      self->filenamefree=1;
+    }
 
   /* type checking */
   if(strlen(vfs)==0)
@@ -1885,27 +1916,6 @@ APSWVFSFile_init(APSWVFSFile *self, PyObject *args, PyObject *kwds)
   file=PyMem_Malloc(vfstouse->szOsFile);
   if(!file) goto finally;
 
-  if(utf8name!=Py_None)
-    {
-      /* The filename has to be passed through xFullPathname.  That
-	 may or may not already have happened, but we do it a second
-	 time to be sure. */
-      int fpres;
-      APSW_FAULT_INJECT(vfsfileopen_fullpathnamemallocfailed,
-			self->filename=PyMem_Malloc(vfstouse->mxPathname+1),
-			self->filename=(char*)PyErr_NoMemory());
-      if(!self->filename)
-	goto finally;
-      APSW_FAULT_INJECT(vfsfileopen_fullpathnamefailed,
-			fpres=vfstouse->xFullPathname(vfstouse, PyBytes_AS_STRING(utf8name), vfstouse->mxPathname, self->filename),
-			fpres=SQLITE_NOMEM);
-      if(fpres!=SQLITE_OK)
-	{
-	  SET_EXC(fpres, NULL);
-	  goto finally;
-	}
-    }
-
   xopenresult=vfstouse->xOpen(vfstouse, self->filename,  file, (int)flagsin, &flagsout);
   SET_EXC(xopenresult, NULL);
   if(PyErr_Occurred())
@@ -2335,7 +2345,7 @@ apswvfsfilepy_xSync(APSWVFSFile *self, PyObject *args)
 static int
 apswvfsfile_xSectorSize(sqlite3_file *file)
 {
-  int result=512;
+  int result=4096;
   PyObject *pyresult=NULL;
   FILEPREAMBLE;
 
@@ -2354,7 +2364,7 @@ apswvfsfile_xSectorSize(sqlite3_file *file)
   if(PyErr_Occurred())
     {
       AddTraceBackHere(__FILE__, __LINE__, "apswvfsfile_xSectorSize", NULL);
-      result=512; /* could be -1 as stated above */
+      result=4096; /* could be -1 as stated above */
     }
 
   Py_XDECREF(pyresult);
@@ -2366,13 +2376,13 @@ apswvfsfile_xSectorSize(sqlite3_file *file)
 
     Return the native underlying sector size. SQLite uses the value
     returned in determining the default database page size. If you do
-    not implement the function or have an error then 512 (the SQLite
+    not implement the function or have an error then 4096 (the SQLite
     default) is returned.
 */
 static PyObject *
 apswvfsfilepy_xSectorSize(APSWVFSFile *self)
 {
-  int res=512;
+  int res=4096;
 
   CHECKVFSFILEPY;
   VFSFILENOTIMPLEMENTED(xSectorSize, 1);
@@ -2833,3 +2843,154 @@ static PyTypeObject APSWVFSFileType =
     APSW_PYTYPE_VERSION
   };
 
+/** .. class:: URIFilename
+
+    SQLite uses a convoluted method of storing `uri parameters
+    <http://www.sqlite.org/uri.html>`__ after the filename binding the
+    C filename representation and parameters together.  This class
+    encapsulates that binding.  The :ref:`example <example-vfs>` shows
+    usage of this class.  
+
+    Your :meth:`VFS.xOpen` method will generally be passed one of
+    these instead of a string as the filename if the URI flag was used
+    or the main database flag is set.
+
+    You can safely pass it on to the :class:`VFSFile` constructor
+    which knows how to get the name back out.
+*/
+
+
+/** .. method:: filename() -> str
+    
+    Returns the filename.
+*/
+static PyObject*
+apswurifilename_filename(APSWURIFilename *self)
+{
+  return convertutf8string(self->filename);
+}
+
+/** .. method:: uri_parameter(name) -> str
+
+    Returns the value of parameter `name` or None.
+
+    -* sqlite3_uri_parameter
+*/
+static PyObject*
+apswurifilename_uri_parameter(APSWURIFilename *self, PyObject *param)
+{
+  const char *res;
+  PyObject *asutf8=getutf8string(param);
+  if(!asutf8)
+    return NULL;
+  res=sqlite3_uri_parameter(self->filename, PyBytes_AS_STRING(asutf8));
+  Py_DECREF(asutf8);
+  return convertutf8string(res);
+}
+
+/** .. method:: uri_int(name, default) -> int
+
+    Returns the integer value for parameter `name` or `default` if not
+    present.
+    
+    -* sqlite3_uri_int64
+*/
+static PyObject*
+apswurifilename_uri_int(APSWURIFilename *self, PyObject *args)
+{
+  char *param=NULL;
+  long long res=0;
+
+  if(!PyArg_ParseTuple(args, "esL", STRENCODING, &param, &res))
+    return NULL;
+  
+  res=sqlite3_uri_int64(self->filename, param, res);
+  PyMem_Free(param);
+
+  return PyLong_FromLongLong(res);
+}
+
+/** .. method:: uri_boolean(name, default) -> bool
+    
+    Returns the boolean value for parameter `name` or `default` if not
+    present.
+
+    -* sqlite3_uri_boolean
+ */
+static PyObject*
+apswurifilename_uri_boolean(APSWURIFilename *self, PyObject *args)
+{
+  char *param=NULL;
+  int res=0;
+
+  if(!PyArg_ParseTuple(args, "esi", STRENCODING, &param, &res))
+    return NULL;
+  
+  res=sqlite3_uri_boolean(self->filename, param, res);
+  PyMem_Free(param);
+
+  if(res)
+    Py_RETURN_TRUE;
+  Py_RETURN_FALSE;
+}
+
+
+static PyMethodDef APSWURIFilenameMethods[]={
+  {"filename", (PyCFunction) apswurifilename_filename, METH_NOARGS, "Get filename"},
+  {"uri_parameter", (PyCFunction) apswurifilename_uri_parameter, METH_O, "Get URI value"},
+  {"uri_int", (PyCFunction) apswurifilename_uri_int, METH_VARARGS, "Get URI integer value"},
+  {"uri_boolean", (PyCFunction) apswurifilename_uri_boolean, METH_VARARGS, "Get URI boolean value"},
+  /* Sentinel */
+  {0, 0, 0, 0}
+};
+
+static PyTypeObject APSWURIFilenameType=
+  {
+    APSW_PYTYPE_INIT
+    "apsw.URIFilename",        /*tp_name*/
+    sizeof(APSWURIFilename),   /*tp_basicsize*/
+    0,                         /*tp_itemsize*/
+    0,                         /*tp_dealloc*/ 
+    0,                         /*tp_print*/
+    0,                         /*tp_getattr*/
+    0,                         /*tp_setattr*/
+    0,                         /*tp_compare*/
+    0,                         /*tp_repr*/
+    0,                         /*tp_as_number*/
+    0,                         /*tp_as_sequence*/
+    0,                         /*tp_as_mapping*/
+    0,                         /*tp_hash */
+    0,                         /*tp_call*/
+    0,                         /*tp_str*/
+    0,                         /*tp_getattro*/
+    0,                         /*tp_setattro*/
+    0,                         /*tp_as_buffer*/
+    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_HAVE_VERSION_TAG, /*tp_flags*/
+    "Filename and URI",        /* tp_doc */
+    0,		               /* tp_traverse */
+    0,		               /* tp_clear */
+    0,		               /* tp_richcompare */
+    0,		               /* tp_weaklistoffset */
+    0,		               /* tp_iter */
+    0,		               /* tp_iternext */
+    APSWURIFilenameMethods,    /* tp_methods */
+    0,                         /* tp_members */
+    0,                         /* tp_getset */
+    0,                         /* tp_base */
+    0,                         /* tp_dict */
+    0,                         /* tp_descr_get */
+    0,                         /* tp_descr_set */
+    0,                         /* tp_dictoffset */
+    0,                         /* tp_init */
+    0,                         /* tp_alloc */
+    0,                         /* tp_new */
+    0,                         /* tp_free */
+    0,                         /* tp_is_gc */
+    0,                         /* tp_bases */
+    0,                         /* tp_mro */
+    0,                         /* tp_cache */
+    0,                         /* tp_subclasses */
+    0,                         /* tp_weaklist */
+    0                          /* tp_del */
+    APSW_PYTYPE_VERSION
+  };
